# read_bragg_reflections_from_out
import re, pandas as pd, math
from io import StringIO


class BraggContainer:
  def __init__(self, outfilepath, wavelength):
    """ searches an out file for bragg reflections generated by FullProf

    returns:
      dictionary:
        keys: phase names
        values: pd.DataFrame with information generated by FullProf

    """


    """ this indicator show when to start collecting the bragg reflection information from the out file """
    indicator = "--------------------------------------------------------------------------------------------------------------"
    """ this indicator showns when we terminate the search """
    indicator2 = "     -----------------------------------------------------"
    header = "No.  Code     H   K   L  Mult     Hw         2theta      Icalc       Iobs      Sigma      HwG     HwL       ETA       d-hkl        CORR"
    phases = []
    self.dfs = dict()
    lines = []
    phase_num = 0
    if_first = True
    stop_all = False
    with open(outfilepath,"r") as reading:
      for line in reading:
        if stop_all:
          break
        if indicator in line:
          if if_first:
            phase = reading.__next__()
            if_first = False
            phases.append( phase )
          # clean phase
          line=reading.__next__()
          line=reading.__next__()
          lines.append("")
          while True:
            try:
              line = reading.__next__()
            except StopIteration:
              break
            if header in line:
              reading.__next__()
              continue
            elif indicator in line:
              phase = reading.__next__()
              phases.append( phase )
              break
            elif indicator2 in line:
              stop_all = True
              break
            lines[-1]+=line

    for phase,line in zip(phases,lines):
      phase = phase.strip().split(" Phase name: ")[-1]
      load = StringIO(line)
      self.dfs[phase] = pd.read_csv(load, sep="\s+", names = header.strip().split())
      self.dfs[phase]["q-space"] = math.pi * 4 / wavelength * ( (math.pi / 180 / 2 * self.dfs[phase]["2theta"]).apply(math.sin) )



# __init__.py

import os, sys
paths_to_append = []
paths_to_append.append( os.path.join(os.path.dirname(__file__), "modules","_plot_seq") )
paths_to_append.append( os.path.join(os.path.dirname(__file__), "modules","_plot_prf") )
paths_to_append.append( os.path.join(os.path.dirname(__file__), "modules") )
paths_to_append.append( os.path.dirname(__file__) )
for p in paths_to_append:
  if p not in sys.path:
    """ if we do not check, we end up with copies of same paths in sys.path """
    sys.path.append( p )

del paths_to_append # try if this parameter does not show up in dir(magnetmatter)?
from plot_prf import plot_prf
from plot_seq import plot_seq
from auxiliary import natural_keys



# extract_pcr.py

def getwavelength(pcrpath):
  """ searches .pcr file for wavelength

  example snippet from pcrpath:
    ! Lambda1  Lambda2    Ratio    Bkpos    Wdt    Cthm     muR   AsyLim   Rpolarz  2nd-muR -> Patt# 1
    0.207000 0.207000  1.00000   40.000 25.0000  0.9999  0.0000    0.00    0.0000  0.0000

  """
  with open(pcrpath, "r") as reading:
    for line in reading:
      if "Lambda1  Lambda2    Ratio" in line:
        line = next(reading)
        w1, w2, ratio = [float(item) for item in line.split()[:3]]
        wavelength = (float(w1) + float(w2)*float(ratio)) / (1+float(ratio))
        return wavelength
  return 0

def getphases(pcrpath):
  """ searches .pcr file for phases.

  example snippet from pcrpath:
    !-------------------------------------------------------------------------------
    !  Data for PHASE number:   2  ==> Current R_Bragg for Pattern#  1:   110.90
    !-------------------------------------------------------------------------------
    Fe$_3$O$_4$

  returns:
    list with phases in ordered sequence"""
  with open(pcrpath, "r") as reading:
    retlist = []
    for line in reading:
      if "Data for PHASE number:" in line:
        line = next(reading)
        line = next(reading).strip()
        retlist.append(line)
    return retlist
  return 0



# datfiles2csv.py

import os, numpy as np, pandas as pd, sys, re
import numpy as np, pandas as pd, re, os
sys.path.append(r"C:\AU-PHD\General_Data\_python\PyPi\magnetmatter\magnetmatter\modules")
from auxiliary import natural_keys
from wrappers import time_response





@time_response
def genereate_CSV_from_dioptas_dat(workingdir = ""):
    """ generate CSV from dioptas .dat files.

    last edit: 2018-01-31

    Function is written to extract 2th values and Intensities from
    2D plots transformed to line diffractograms by the software Dioptas.

    OUTPUT:
        Saves a pd.DataFrame to "mycsv.csv" at "path/python_output".
    NOTE:
        if two frames have different number of angles, NaNs will be present.
        NaNs are replaced with zeroes.
    Note to Pelle:
        DO NOT TRY TO EXTRACT ERRORS. THERE ARE NONE GENERATED FROM DIOPTAS.
    """
    workingdir = workingdir if workingdir != "" else os.getcwd()
    os.chdir(workingdir)

    """get all dat files"""
    files = [d for d in os.listdir() if d.endswith(".dat")]

    """ skipping treatment if no dat files are found """
    if len(files) == 0:
      return pd.DataFrame

    """sorting the frames"""
    files.sort(key = natural_keys)

    """ only take every skipping'th frame to plot
    the higher number for skipping, the faster (but cruder)
    the visualization of the data frames will be.

    EDIT: removed... well left for rapid testing..."""
    # files = files[::100]

    """extacting 2th and counts"""
    df = 0
    index = "angle"
    i = 0; imax = len(files)

    """ loops through all .dat files"""
    for dat in files:
      i += 1
      # if i-1 == imax:
      #     break
      frame = dat.strip(r".dat")
      """friendly output to user"""
      if i % 100 == 0:
        print(i, "of", imax,". df.shape =",df.shape)

      with open(dat) as f:
        """loads in data and merge on angles"""

        """ the dioptas generated dat files are created with numpy's
        .savetxt function. Therefore loadtxt is used without hesitation """
        loaded = np.loadtxt(f)

        """ transforming loaded numpy array to pandas dataframe """
        df_tmp = pd.DataFrame(loaded, columns = [index, frame])

        """ checking if df is a pandas.DataFrame """
        if type(df) != type(df_tmp):
          df = df_tmp
        else:
          """ merges dataframes by filling nonmatching dimensions with NaNs """
          df = df.merge(df_tmp, on = index, how = "outer")

    """ fills all NaN with zeroes """
    df = df.fillna(0)

    """ saves pd.DataFrame to .csv"""
    df.to_csv(os.path.join(os.path.join("..","_dioptas_dats.csv")), index = False)

    return df






def generate_CSV_from_dmc_dat(workingdir = ""):
  """Extract predata from .dat files (DMC instrument)
  Args:
    path: full path to folder with .dat files.
  Returns:
    pandas.DataFrame with columns = timestamps, index = 2th angles, values = extracted signal values.
  Note:
    the errors are also extracted, but not returned in the current code.
  """
  oldpath = os.getcwd()
  os.chdir(workingdir)
  dats = [dat for dat in os.listdir() if dat.endswith(".dat")]
  if len(dats) == 0:
    print("skipping ",os.getcwd())
    return -1
  """return int if char is a digit, otherwise a string char is returned"""
  # return_chars_or_digits = lambda text: int(text) if text.isdigit() else text
  '''alist.sort(key=natural_keys) sorts in human order'''
  # natural_keys = lambda text: [return_chars_or_digits(c) for c in re.split("(\d+)", text)]
  """sorting the frames"""
  dats.sort(key=natural_keys)

  firstRun = True
  frames = []
  for num, file in enumerate(dats):
    frames.append(num+1) #float(file.strip("frame").strip(".dat")) )
    with open(file,"r") as fileHandler:
      open_file = fileHandler.read().split("\n")

    gdic = dict()
    for index, line in enumerate(open_file):
      # print(line)
      for linei in line.split(","):
        for i in linei.split(";"):
          g = re.search("(.*)=(.*)",i)
          if g:
            gdic[g.group(1).strip()] = g.group(2).strip("'")
    metadata1 = open_file[:3]
    data = open_file[3:-7]
    if firstRun:
      for i, line in enumerate(open_file):
        if "Filelist" in line:
          startmetadata2 = i
          break
    metadata2 = open_file[startmetadata2:]
    halfway = int(len(data) / 2)
    signal = data[ : - halfway ]
    error =  data[ halfway : ]

    signal = [s.split() for s in signal]
    error = [s.split() for s in error]

    flatten = lambda l: [float(item) for sublist in l for item in sublist]

    signal = flatten(signal)
    error = flatten(error)

    metadata1 = [line.split(",") for line in metadata1]
    metadata2 = [line.split() for line in metadata2]

    sample_description = metadata1[0][1]
    wavelength = metadata1[1][0]
    date = metadata1[1][-1]
    angle = metadata1[2][0]
    sample = metadata1[2][1]
    frame_number = metadata2[0]

    metadata = [sample_description, wavelength, date, sample, frame_number]

    if firstRun:
      # angle = [float(i) for i in angle.split()]
      angles = [int(float(i)*10) for i in angle.split()]
      low2th, step2th, high2th, maxcounts = angles
      # break
      angles = [float(i)/10 for i in range( low2th, high2th + 1, step2th) ] # +1 for pythons range system
      angles = np.array(angles)
      signals = np.ndarray(shape = ( len(dats), len(signal) ), dtype="float64")
      errors = np.ndarray(shape = ( len(dats), len(error) ), dtype="float64")
      # metadatas = np.ndarray(shape = ( len(dats), len(metadata) ), dtype="s20")
      metadatas = []
      firstRun = False
    signals[num] = signal
    errors[num] = error
    metadatas.append(metadata)
  import pandas as pd
  from datetime import datetime
  import datetime as dt
  frames = [" ".join(["frame",str(int(num))]) for num in frames]
  df = pd.DataFrame(data=signals.T, index=angles, columns = frames)
  df.index.name = "angle"

  """extracting true frame and timestamps from .dat files"""
  dates = list(); frames = list()
  for dat in dats:
    with open(dat, "r") as datfile:
      for line in datfile:
        found1 = re.search("Date=\'([\d\s\-:]+).*\'",line) # Date='2016-12-21 17:04:35'
        found2 = re.search("Filelist=\'.*:\d*:(\d+-\d+)\'",line) # Filelist='dmc:2016:91790-91793'
        if found1:
          date = found1.group(1)
          date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
          dates.append(date)
          continue
        elif found2:
          frame = found2.group(1)
          frames.append(frame)
          break
  # + 8 to accomodate that it took 8 minutes to acquire first frame.
  # first_frame_timedelta = dt.timedelta(minutes = 8)
  deltatimes = [date - dates[0] for date in dates]
  seconds = [dt.seconds for dt in deltatimes]
  minutes = [second / 60.0 for second in seconds]
  df.columns = minutes
  # import pdb; pdb.set_trace();
  df.to_csv(os.path.join(os.path.join("..","_dmc_dats.csv")), index = True)
  os.chdir(oldpath)
  return df
#
#
#


















  # end




import os, pandas as pd, sys, math, re
import matplotlib.pyplot as plt

""" importing private modules """
from find_tools import findfile, findfolders_abspath
from read_prf import read_prf
from extract_pcr import getwavelength, getphases
from auxiliary import cm2inch, natural_keys
from read_bragg_reflections_from_out import BraggContainer
import numpy as np





class AbstractPrf():
  """ we will collect the many lose variables...

  this is designed to be a abstract class. I have not yet figured out the benefits
  of using the abc module to do this. for now I use built-in inheritate for PlainPrf and EnhancedPrf. """
  def __init__(self, path = "", dpi = 100, verbose = False):

    self.dpi = dpi
    """ path bookkeeping """
    self.basepath = path if path != "" else os.getcwd()
    os.chdir(self.basepath)
    self.verbose = verbose
    self.basefolders = findfolders_abspath()
    if self.verbose:
      for f in self.basefolders: print("found folder {}".format(f))
      print("")
      """ TODO:
      read in all data and metadata to keep (save reading in files over and over).
      Through away folders in basefolders that cannot be plotted. """
    # self.basefolders = iter( self.basefolders ) # no longer needed.

  def update(self):

    # verbose = verbose if verbose != "" else self.verbose

    if self.verbose: print("\nworking on", self.basefolder)

    os.chdir(self.basefolder)

    def helper_findfile(extension):
      myfile = findfile(extension)
      if not os.path.exists(myfile):
        if self.verbose: print("!-!-!-! skipping \"" + self.basefolder + "\" + --> no " + extension + " was found");
        return -1
      if extension == ".prf":
        self.baseprf = myfile;
      if extension == ".pcr":
        self.basepcr = myfile;
      if extension == ".out":
        self.baseout = myfile;
      return 0

    if helper_findfile(".prf") == -1: return -1
    if helper_findfile(".pcr") == -1: return -1
    if helper_findfile(".out") == -1: return -1

    """ reading DMC/HRPT or Xray/Dioptas prf file """
    (prf_type, self.df) = read_prf(self.baseprf)
    if self.verbose: print("found {}-prf.".format(prf_type))
    should_I_break = [x for x in ["dmc", "xray"] if x == prf_type]
    if len(should_I_break) == 0:
      return -1

    """ check if model data are actual values! otherwise skip datafolder """
    if self.df.empty:
      if self.verbose: sys.stdout.write("!-!-!-! skipping folder \"" + self.basefolder + "\". DataFrame is empty (values in prf might be NaNs)\n");
      return -1

    self.wavelength = getwavelength(self.basepcr)
    self.df["Yobs-Ycal"] = self.df["Yobs"]-self.df["Ycal"]
    """ formular for converting from 2Theta to Q-space:
    |q| = (4 * pi) / wavelength * sin(2Theta[radians] / 2)
    source: https://www.staff.tugraz.at/manfred.kriechbaum/xitami/java/qtheta.html """
    self.df["q-space"] = math.pi * 4 / self.wavelength * ( (math.pi / 180 / 2 * self.df["2Theta"]).apply(math.sin) )

    return 0

  def baseplotting(self):
    plt.close("all")
    self.basefig, self.baseax = plt.subplots(figsize = cm2inch(self.graph_width, self.graph_height) )

    if self.xlim != None:
      df = self.df[self.df["q-space"] > self.xlim[0]]
      df = df[df["q-space"] < self.xlim[1]]
    else:
      df = self.df

    if df.empty: return -1

    y = "Yobs"
    self.baseax = df.plot(
      "q-space",
      y,
      kind="scatter",
      s = self.dotsize,
      ax = self.baseax,
      fontsize = self.smaller_text,  # xtick and ytick labels
      label = y, # for legend
      c = "r",
      alpha=0.5,
      )
    df.plot(
      "q-space",
      "Ycal",
      linewidth = self.linewidth,
      ax = self.baseax,
      alpha=0.8,
      )
    df.plot(
      "q-space",
      "Yobs-Ycal",
      linewidth = self.linewidth,
      ax = self.baseax,
      alpha=0.3,
      )

    return 0

  def update_legends(self):
    """ changing the position of the legends to make Yobs appear at the top """
    handles,labels = self.baseax.get_legend_handles_labels()
    handles = [handles[2], handles[0], handles[1]]
    labels = [labels[2], labels[0], labels[1]]

    """ setting legend to best position """
    self.legend = plt.legend(handles,labels, loc="best", prop = {"size": self.micro_text})

  def update_xylabels_tightlauoyt(self):
    """ setting dummy y- and xlabel names. These are changed later
    self is done due to the effect of using plt.tight_layout()
    in combination with rotation and shifting of the text object"""
    ylabel = plt.ylabel("Counts", size = self.smaller_text)
    plt.xlabel(r"2$\theta$", size = self.smaller_text)

    """ repositioning and rotation the ylabel """
    self.baseax.yaxis.set_label_coords(*self.ylabelpos_pre)
    ylabel.set_rotation(0)

    """ Xlabel must be here otherwise plt.tight_layout is not given the
    additional white space for the ref_info text """
    self.baseax.xaxis.set_label_coords(*self.xlabelpos_pre)

    """ set all yticks to 100 to tweak the right space for final ytick labels """
    self.myyticks = self.baseax.get_yticks()
    self.mynewyticks = [self.yticklabel_placeholder for x in self.myyticks]
    plt.yticks(self.myyticks, self.mynewyticks)

    """ ensure nothing is left outside the plot
    many elements are moved the following code to optimize the space """
    plt.tight_layout()
    """ everything from here is tweaking already existing plt objects! """

    """ setting padding of x and y tick labels """
    self.baseax.tick_params(pad=self.ticklabelpadding)

    """ repositioning and renaming xlabel and ylabel
    _after_ the use of plt.tight_layout()"""
    ylabel.set_rotation(90)
    self.baseax.yaxis.set_label_coords(*self.ylabelpos_post)
    self.baseax.xaxis.set_label_coords(*self.xlabelpos_post)
    # plt.xlabel(r"scattering angle 2$\theta$ [$^o$]", size = smaller_text)
    plt.xlabel(r"4$\pi$/$\lambda\cdot$sin($\theta$) = Q [$\AA^{-1}$]", size = self.smaller_text)
    ylabel = plt.ylabel("signal count [a.u.]", size = self.smaller_text)

  def update_title(self):
    def _label_text(mystring):
      """ returns a latexified string to be used in plt.
      e.g. 500oC --> 500$^o$C """
      return re.sub("(\d)oC","\g<1>°C",mystring)
    self.basetitle = _label_text(os.path.basename(self.basefolder))
    plt.title(self.basetitle, size = self.medium_text)

  def update_xylim(self, xlim):
    """ making the diffraction data as wide as possible """
    # xmin = min(self.df["q-space"])
    # xmax = max(self.df["q-space"])
    # self.df = self.df[self.df["Yobs"] > 2.8]
    # self.df = self.df[self.df["Yobs"] < 3.6]

    """ this is not really necessary with plt.xlim?? """
    df = self.df[self.df["q-space"] > xlim[0]]
    df = df[df["q-space"] < xlim[1]]
    ylim = (df[["Yobs","Ycal","Yobs-Ycal"]].min().min(), df[["Yobs","Ycal","Yobs-Ycal"]].max().max())
    self.baseax.set_ylim( *ylim )
    self.baseax.set_xlim( *xlim )
    """ this is not really necessary with plt.xlim?? """


    # ymin = min([min(self.df["Yobs"]), min(self.df["Ycal"])])
    # ymax = max([max(self.df["Yobs"]), max(self.df["Ycal"])])
    # # plt.xlim(1,3.5)
    # """ reducing empty space in data plot """
    # min_y = min(self.baseax.get_yaxis().major.locator.locs)
    # max_y = max(self.baseax.get_yaxis().major.locator.locs)
    # plt.ylim(min_y,max_y)
    # plt.ylim(ymin,ymax)
    # plt.xlim(2.8,3.6)


  def update_xyticks(self):
    """ showing only first and last ytick label """
    self.mynewyticks = ["" for x in self.myyticks]
    self.mynewyticks[-1] = "{:.1e}".format(self.myyticks[-1])
    self.mynewyticks[0] = "{:.1e}".format(self.myyticks[0])
    plt.yticks(self.myyticks, self.mynewyticks)

    myxticks = self.baseax.get_xticks()
    mynewxticks = []
    for x in myxticks:
      mynewxticks.append(round(x,1))
    mynewxticks[-1] = ""
    mynewxticks[0] = ""
    plt.xticks(myxticks, mynewxticks)

  def update_grid(self):
    self.baseax.grid()


  def savefigure(self):
    """ saving the figure in high resolution """
    """ path and name of output file """

    plt.savefig(self.figname, dpi=self.dpi, format = "png")

    """ avoid memory leak """
    plt.close()

  def add_bragg_peaks(self):
    bc = BraggContainer(self.baseout, self.wavelength).dfs

    df2 = self.df
    if self.xlim != None:
      df2 = df2[df2["q-space"] > self.xlim[0]]
      df2 = df2[df2["q-space"] < self.xlim[1]]
      if df2.empty: return -1

    mininum = min( self.baseax.get_ylim() )
    maximum = df2["Yobs"].min()

    myrange = iter(np.linspace(0, mininum, 10)+maximum)
    next(myrange)
    phases = getphases(self.basepcr)
    for phase in sorted( phases , key = natural_keys):
      minmax = (next(myrange), next(myrange))
      df = bc[phase]

      if self.xlim != None:
        df = df[df["q-space"] > self.xlim[0]]
        df = df[df["q-space"] < self.xlim[1]]

      if df.empty: continue

      # reflections = df["2theta"].unique()
      reflections = df["q-space"].unique()
      for refl in reflections:
        # print(phase, refl, *minmax)
        self.baseax.plot([refl, refl],[*minmax],"--r", linewidth = 0.5, alpha = 1, zorder=1)







  def add_bragg_peaks_obsolete_remove_soon(self):
    """ uPDaTe:
    we might be able to extract bragg peaks
    from the FP .outfile
    """

    """ inserting bragg peak positions.

    NOTE: the 2theta values can be found in the .out file!!!!!!!!!

    No.  Code     H   K   L  Mult     Hw         2theta      Icalc       Iobs      Sigma      HwG     HwL       ETA       d-hkl        CORR

    NOTE: implementation is not correct! see discussion below
    """
    type_of_diffraction_data = ""
    if type_of_diffraction_data == "more complicated than as such to implement":
      import bragg_peak_indications as bpi

      mycs = bpi.Crystals().crystals
      mypositions = [0.02, 0.04]
      mycolours = ["r","b","g"]*2
      for myc in mycs:
        mycolour = mycolours.pop(0)
        myc = myc.unique
        mymax = max(df["2Theta"])
        mymin = min(df["2Theta"])
        myc = myc[myc > mymin]
        myc = myc[myc < mymax]
        for val in myc:

          setmin = mypositions[0] * (abs(min_y) + max_y) + min_y
          setmax = mypositions[1] * (abs(min_y) + max_y) + min_y
          plt.plot([val,val],[setmin, setmax], c = mycolour, lw = 0.5)
        mypositions = [0.03 + mypos for mypos in mypositions]
    """
    above code is not correct - magnetic structures are inherently complicated
    and even for x-ray diffraction patterns the zero shift (if not more...) has to be taken into account.
    Also I tried to color coordinate the phasename with the reflections, but changing color in a block of
    text by means of LaTeX-based code has proven to be non-trivial.
    """



# plot_prf.py

import os, numpy as np, pandas as pd, sys, re, math
import matplotlib.pyplot as plt

""" importing private modules """
from auxiliary import cm2inch
from wrappers import TraceCalls
from find_tools import findfile, findfiles, findfolders_abspath
from prf_plot_tools import extract_out_file, calculate_size, excluded_regions
from prf_plot_tools import prepare_refinement_result, get_spacegroups
from read_prf import read_prf
from extract_pcr import getwavelength
from myObject import AbstractPrf


def plot_prf(path = "", printsize = "two_in_docx", video = False, xlim = (2.8,3.6), dpi = 250):
  # plot_prf_enhanced(path=path, printsize=printsize, minmax2theta = minmax2theta)
  # plot_prf_plain(path=path, printsize=printsize, minmax2theta = minmax2theta)
  plain = PlainPrf(printsize = printsize, path = path, verbose = False, dpi = dpi)
  enhanced = EnhancedPrf(printsize = printsize, path = path, verbose = True, dpi = dpi)


  enhanced( xlim = xlim )
  plain( xlim = None )
  xlimrange = True
  if video:
    from make_video import Video
    xlim = (0.0,16.0)
    range1 = [round(x,1) for x in np.arange(*xlim,0.1)]
    range2 = [round(x+0.4,1) for x in np.arange(*xlim,0.1)]
    print("")
    mymax = max(range2)
    mymin = max(range1)
    for low, high in zip(range1, range2):
      # import pdb; pdb.set_trace();
      sys.stdout.write("\rplotting q-range = ({},{})  ... continues to ({},{})         ".format(low,high,mymin,mymax))
      plain( xlim = (low, high))
    print("")
    Video(plain.outfolder)
  else:
    plain( xlim = xlim )



class PlainPrf(AbstractPrf):
  def __init__(self, printsize = "two_in_docx", path = "", verbose = False, dpi = 100):
    super().__init__(path = path, verbose = verbose, dpi = dpi)
    if printsize == "two_in_docx":
      self.graph_width = 8.45 # cm
      self.graph_height = self.graph_width*3/4 # cm
      self.medium_text = 8 # still quite big
      self.smaller_text = 6 # better
      self.micro_text = 4 # 4 good for refinement info.
      self.dotsize = 1 #
      self.linewidth = 0.7
      self.ylabelpos_pre = (0.01,1.05)
      self.ylabelpos_post = (-0.058,0.50)
      self.xlabelpos_pre = (1.05,0.15)
      self.xlabelpos_post = (0.50,-0.10)
      self.ticklabelpadding = 3
      self.yticklabel_placeholder = "10000"
    elif printsize == "one_in_docx":
      self.graph_width = 17 # cm
      self.graph_height = self.graph_width*3/4 # cm
      self.medium_text = 12 # still quite big
      self.smaller_text = 10 # better
      self.micro_text = 8 # 4 good for refinement info.
      self.dotsize = 2
      self.linewidth = 2
      self.ylabelpos_pre = (-0.007,1.05)
      self.ylabelpos_post = (-0.07,0.50)
      self.xlabelpos_pre = (1.05,-0.043)
      self.xlabelpos_post = (0.50,-0.07)
      self.ticklabelpadding = 5
      self.yticklabel_placeholder = "100"
    elif printsize == "one_in_ppt":
      self.graph_width = 32 # cm
      self.graph_height = self.graph_width*2/5 # cm
      self.medium_text = 18 # still quite big
      self.smaller_text = 16 # better
      self.micro_text = 10 # 4 good for refinement info.
      self.dotsize = 3
      self.linewidth = 2
      self.ylabelpos_pre = (-0.009,1.07)
      self.ylabelpos_post = (-0.045,0.55)
      self.xlabelpos_pre = (1.05,-0.063)
      self.xlabelpos_post = (0.50,-0.10)
      self.ticklabelpadding = 9
      self.yticklabel_placeholder = "1000000"
    elif printsize == "one_in_ppt":
      self.graph_width = 32 # cm
      self.graph_height = self.graph_width*2/5 # cm
      self.medium_text = 18 # still quite big
      self.smaller_text = 16 # better
      self.micro_text = 10 # 4 good for refinement info.
      self.dotsize = 3
      self.linewidth = 2
      self.ylabelpos_pre = (-0.009,1.07)
      self.ylabelpos_post = (-0.045,0.55)
      self.xlabelpos_pre = (1.05,-0.063)
      self.xlabelpos_post = (0.50,-0.10)
      self.ticklabelpadding = 9
      self.yticklabel_placeholder = "1000000"



    plt.style.use("seaborn-white")

  def update_outpath(self):
    self.suffix = "_plain"
    self.outfolder = os.path.join(self.basepath, self.suffix)
    if not os.path.exists(self.outfolder): os.makedirs(self.outfolder)
    out = os.path.join(self.outfolder, os.path.basename(self.basefolder))
    if self.xlim != None:
      xlim = tuple(str(round(x,1)) for x in self.xlim)
      self.suffix += xlim[0]+"to"+xlim[1]+".png"
    else:
      out = os.path.join(self.basepath, os.path.basename(self.basefolder))
      self.suffix += ".png"
    self.figname = out + self.suffix

  def update_legends(self):
    super(PlainPrf, self).update_legends()
    self.legend.set_visible(False)

# @TraceCalls()
  def __call__(self, xlim = ""):
      """
      ?searches all folders for given path for .prf, .out, .pcr files to plot with refined informations?

      ?Args?:
          path: script will be executed at this path
          printsize: options are "two_in_docx", two_in_ppt", "one_in_docx", "one_in_ppt".
          minmax2theta: tuple of min and max 2theta values to be ignored

      ?Compatibility?:
          Script has been tested with refined data from DMC, HRPT (SINQ:PSI)
          and refined data from X-ray sources.
          The FullProf .pcr format is the new format (_not_ classical).

      ?Side effect?:
          Saves plots of "Yobs", "Ycal" and "Ycal-Yobs" for .prf, .out and .pcr files
          found in child folderns. The script is verbose and will tell what it is currently
          doing, if the process was successful and what mistakes occured.

      ?Common mistake?:
          If you change your .pcr, remember to run FullProf to see the effect of the updated
          parameters or added phases. This script stupidly reads in the .out and .prf files.
          Any changes to especially phase names will certainly result in unchanged phase names found
          in plot.
          Therefore always run the .pcr file to update the FullProf output files.
      """

      self.xlim = xlim

      for self.basefolder in self.basefolders:

        # if not os.path.exists(self.basefolder): continue # not sure if this is necessary

        if 0 != self.update(): continue

        # """ shortening the view of the diffractogram
        #   if user has provided a tuple of min-max values """
        # if type(xlim) == type(tuple()) and len(xlim) == 2:
        #   self.df = self.df[self.df["2Theta"] > minmax2theta[0]]
        #   self.df = self.df[self.df["2Theta"] < minmax2theta[1]]


        """ plotting Yobs, Ycal, and Yobs-Ycal data. """
        if self.baseplotting() == -1: continue
        """ above should be continue... if diffraction data from different sources are used. """

        self.update_legends()

        self.update_title()

        self.add_bragg_peaks()


        self.update_xylabels_tightlauoyt()

        self.update_xyticks()

        if self.xlim != None:
          self.baseax.set_xlim( *self.xlim )
        else:
          self.baseax.set_xlim( self.df["q-space"].min(), self.df["q-space"].max() )


        """ THIS DOESN'T work. ups capslock... """
        plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='on',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        # labelbottom='off'
        ) # labels along the bottom edge are off


        # self.update_grid()
        # if type(xlim) == type(tuple()): self.update_xylim( xlim )



        self.update_outpath()

        self.savefigure()










class EnhancedPrf(AbstractPrf):
  def __init__(self, printsize = "two_in_docx", path = "", verbose = False, dpi = 100):
    super().__init__(path = path, verbose = verbose, dpi = dpi)
    if printsize == "two_in_docx":
      self.graph_width = 8.45 # cm
      self.graph_height = self.graph_width*3/4 # cm
      self.medium_text = 8 # still quite big
      self.smaller_text = 6 # better
      self.micro_text = 4 # 4 good for refinement info.
      self.refsize_text = 5
      self.dotsize = 1 #
      self.linewidth = 0.7
      self.subplot_right_space = 8.1#6.8
      self.refinfo_pos_long_text = (1.02,-0.2)#(1.02,-0.07)
      self.refinfo_pos_short_text = (1.02,-0.07)
      self.ylabelpos_pre = (0.01,1.05)
      self.ylabelpos_post = (-0.11,0.50)
      self.xlabelpos_pre = (1.05,0.15)
      self.xlabelpos_post = (0.50,-0.09)
      self.ticklabelpadding = 3
      self.yticklabel_placeholder = "1000"
    if printsize == "two_in_ppt":
      self.graph_width = 15 # cm
      self.graph_height = self.graph_width*3/4 # cm
      self.medium_text = 12 # still quite big
      self.smaller_text = 10 # better
      self.micro_text = 8 # 4 good for refinement info.
      self.refsize_text = 8
      self.dotsize = 2
      self.linewidth = 2
      self.subplot_right_space = 7.7
      self.refinfo_pos_long_text = (1.03,-0.14)
      self.refinfo_pos_short_text = (1.03,-0.06)
      self.ylabelpos_pre = (-0.008,1.05)
      self.ylabelpos_post = (-0.10,0.50)
      self.xlabelpos_pre = (1.05,-0.043)
      self.xlabelpos_post = (0.50,-0.08)
      self.ticklabelpadding = 6
      self.yticklabel_placeholder = "100"
    elif printsize == "one_in_docx":
      self.graph_width = 17 # cm
      self.graph_height = self.graph_width*3/4 # cm
      self.medium_text = 12 # still quite big
      self.smaller_text = 10 # better
      self.micro_text = 8 # 4 good for refinement info.
      self.refsize_text = 8
      self.dotsize = 2
      self.linewidth = 2
      self.subplot_right_space = 6.8
      self.refinfo_pos_long_text = (1.015,-0.13)
      self.refinfo_pos_short_text = (1.015,-0.055)
      self.ylabelpos_pre = (-0.007,1.05)
      self.ylabelpos_post = (-0.07,0.50)
      self.xlabelpos_pre = (1.05,-0.043)
      self.xlabelpos_post = (0.50,-0.07)
      self.ticklabelpadding = 5
      self.yticklabel_placeholder = "100"
    elif printsize == "one_in_ppt":
      self.graph_width = 32 # cm
      self.graph_height = self.graph_width*2/5 # cm
      self.medium_text = 18 # still quite big
      self.smaller_text = 16 # better
      self.micro_text = 10 # 4 good for refinement info.
      self.refsize_text = 10
      self.dotsize = 3
      self.linewidth = 2
      self.subplot_right_space = 4.6
      self.refinfo_pos_long_text = (1.01,-0.2)
      self.refinfo_pos_short_text = (1.01,-0.075)
      self.ylabelpos_pre = (-0.009,1.07)
      self.ylabelpos_post = (-0.045,0.55)
      self.xlabelpos_pre = (1.05,-0.063)
      self.xlabelpos_post = (0.50,-0.10)
      self.ticklabelpadding = 9
      self.yticklabel_placeholder = "1000000"

    plt.style.use("seaborn-white")

  def update_outpath(self):
    out = os.path.join(self.basepath, os.path.basename(self.basefolder) )
    suffix = ""
    if self.xlim != None:
      xlim = tuple(str(round(x,1)) for x in self.xlim)
      suffix += xlim[0]+"to"+xlim[1]
    self.figname = out + ".png"

  def plotting_info_text(self):
    plt.subplots_adjust(right=self.subplot_right_space)
    """ extraction of:
    phase names (dict)
    phase fractions (dict)
    wavelength (float)
    refined parameter value and error (dict) """
    self.wavelength, self.frac_info, self.phases, self.refined_par = extract_out_file()
    """ can we split this method up or something else? """

    """ calculates:
    ab and c-sizes (dict of dicts) """
    self.size_info = calculate_size(
        wavelength = self.wavelength,
        phases = self.phases,
        refined_par = self.refined_par,
        )

    """ extracting spacegroups from PCR file """
    self.spacegroups = get_spacegroups()

    """ preparing information such as refined parameters, phases, phase fractions
    crystallite sizes and so to be printed on canvas. """
    self.combined_text = prepare_refinement_result(
        phases = self.phases,
        refined_par = self.refined_par,
        size_info = self.size_info,
        frac_info = self.frac_info,
        spacegroups = self.spacegroups,
        wavelength = self.wavelength,
        )

    """ if not long refinement info text, aling text with bottom of graph """
    self.refinfo_pos = self.refinfo_pos_short_text if self.combined_text.count("\n") < 24 else self.refinfo_pos_long_text

    """ plotting refinement info on canvas"""
    ttt = self.baseax.text(
        self.refinfo_pos[0],
        self.refinfo_pos[1],
        self.combined_text,
        size=self.refsize_text,
        fontname = "monospace",
        picker=True,
        transform = self.baseax.transAxes,
        # zorder = 0,
        )

    """ making the refinement info have a semitransparent white background """
    # ttt.set_bbox(dict(facecolor='white', alpha=0.5, edgecolor='white'))



  # @TraceCalls()
  def __call__(self,xlim = ""):
      """
      searches all folders for given path for .prf, .out, .pcr files to plot with refined informations

      Args:
          path: script will be executed at this path
          printsize: options are "two_in_docx", two_in_ppt", "one_in_docx", "one_in_ppt".
          minmax2theta: tuple of min and max 2theta values to be ignored

      Compatibility:
          Script has been tested with refined data from DMC, HRPT (SINQ:PSI)
          and refined data from X-ray sources.
          The FullProf .pcr format is the new format (_not_ classical).

      Side effect:
          Saves plots of "Yobs", "Ycal" and "Ycal-Yobs" for .prf, .out and .pcr files
          found in child folderns. The script is verbose and will tell what it is currently
          doing, if the process was successful and what mistakes occured.

      Common mistake:
          If you change your .pcr, remember to run FullProf to see the effect of the updated
          parameters or added phases. This script stupidly reads in the .out and .prf files.
          Any changes to especially phase names will certainly result in unchanged phase names found
          in plot.
          Therefore always run the .pcr file to update the FullProf output files.

      """

      # printsize = "two_in_ppt" # graphs are 17 cm wide, perfect for two in .ppt
      # printsize = "one_in_docx" # graphs are 17 cm wide, perfect for one in word
      # printsize = "one_in_ppt" # single grpah in .ppt
      # printsize = "two_in_docx" # graphs are 8 cm wide, perfect for two in word

      self.xlim = xlim

      for self.basefolder in self.basefolders:

          if 0 != self.update(): continue

          # """ below code is obsolete """
          # if not df[df["2Theta"] > 30.00].empty:
          #   """ if-check is necessary when treating synchrotron data,
          #   where angles are less than 20 degrees! """
          #   # df = df[df["2Theta"] > 20.00]
          #   # df = df[df["2Theta"] < 92.900]

          # """ shortening the view of the diffractogram
          # if user has provided a tuple of min-max values """
          # if type(minmax2theta) == type(tuple()) and len(minmax2theta) == 2:
          #   df = df[df["2Theta"] > minmax2theta[0]]
          #   df = df[df["2Theta"] < minmax2theta[1]]

          if self.baseplotting() == -1: continue

          self.plotting_info_text()

          self.update_legends()

          self.update_title()

          self.add_bragg_peaks()

          self.update_xylabels_tightlauoyt()

          # """ reducing empty space in data plot """
          # min_y = min(self.baseax.get_yaxis().major.locator.locs)
          # max_y = max(self.baseax.get_yaxis().major.locator.locs)
          # plt.ylim(min_y,max_y)

          # if type(xlim) == type(tuple()): self.update_xylim( xlim )

          self.update_xyticks()

          if self.xlim != None:
            self.baseax.set_xlim( *self.xlim )


          self.update_grid()

          self.update_outpath()

          self.savefigure()

      # """ normalizing the path """
      # os.chdir(path)









# print("\n\nEND_OF_SCRIPT")
# print(sys.version_info)
# print(sys.executable)
